@@ -131,15 +131,17 @@ static int cmd_match(char *args[]) {
     const char *pattern = cleanpath(args[0]);
     char **matches;
     int filter = (args[1] != NULL) && (strlen(args[1]) > 0);
+    int result = 0;
 
     cnt = aug_match(aug, pattern, &matches);
-    if (cnt == -1) {
+    if (cnt < 0) {
         printf("  (error matching %s)\n", pattern);
-        return -1;
+        result = -1;
+        goto done;
     }
     if (cnt == 0) {
         printf("  (no matches)\n");
-        return 0;
+        goto done;
     }
 
     for (int i=0; i < cnt; i++) {
@@ -153,10 +155,12 @@ static int cmd_match(char *args[]) {
         } else {
             printf("%s = %s\n", matches[i], val);
         }
-        free((void *) matches[i]);
     }
+ done:
+    for (int i=0; i < cnt; i++)
+        free(matches[i]);
     free(matches);
-    return 0;
+    return result;
 }
 
 static int cmd_rm(char *args[]) {
@@ -190,6 +194,33 @@ static int cmd_set(char *args[]) {
     return r;
 }
 
+static int cmd_defvar(char *args[]) {
+    const char *name = args[0];
+    const char *path = cleanpath(args[1]);
+    int r;
+
+    r = aug_defvar(aug, name, path);
+    if (r == -1)
+        printf ("Failed\n");
+    return r;
+}
+
+static int cmd_defnode(char *args[]) {
+    const char *name = args[0];
+    const char *path = cleanpath(args[1]);
+    const char *value = args[2];
+    int r;
+
+    /* Our simple minded line parser treats non-existant and empty values
+     * the same. We choose to take the empty string to mean NULL */
+    if (value != NULL && strlen(value) == 0)
+        value = NULL;
+    r = aug_defnode(aug, name, path, value, NULL);
+    if (r == -1)
+        printf ("Failed\n");
+    return r;
+}
+
 static int cmd_clear(char *args[]) {
     const char *path = cleanpath(args[0]);
     int r;
@@ -235,6 +266,22 @@ static int cmd_save(ATTRIBUTE_UNUSED char *args[]) {
     return r;
 }
 
+static int cmd_load(ATTRIBUTE_UNUSED char *args[]) {
+    int r;
+    r = aug_load(aug);
+    if (r == -1) {
+        printf("Loading failed\n");
+    } else {
+        r = aug_match(aug, "/augeas/events/saved", NULL);
+        if (r > 0) {
+            printf("Saved %d file(s)\n", r);
+        } else if (r < 0) {
+            printf("Error during match: %d\n", r);
+        }
+    }
+    return r;
+}
+
 static int cmd_ins(char *args[]) {
     const char *label = args[0];
     const char *where = args[1];
@@ -344,7 +391,7 @@ static const struct command const commands[] = {
       "        exist yet. If DST exists already, it and all its descendants are\n"
       "        deleted. If DST does not exist yet, it and all its missing \n"
       "        ancestors are created." },
-    { "set", 2, 2, cmd_set, "set <PATH> <VALUE>",
+    { "set", 1, 2, cmd_set, "set <PATH> <VALUE>",
       "Associate VALUE with PATH. If PATH is not in the tree yet,\n"
       "        it and all its ancestors will be created. These new tree entries\n"
       "        will appear last amongst their siblings"
@@ -369,6 +416,22 @@ static const struct command const commands[] = {
       "Save all pending changes to disk. For now, files are not overwritten.\n"
       "        Instead, new files with extension .augnew are created"
     },
+    { "load", 0, 0, cmd_load, "load",
+      "Load files accordig to the transforms in /augeas/load."
+    },
+    { "defvar", 2, 2, cmd_defvar, "defvar <NAME> <EXPR>",
+      "Define the variable NAME to the result of evalutating EXPR. The\n"
+      "        variable can be used in path expressions as $NAME. Note that EXPR\n"
+      "        is evaluated when the variable is defined, not when it is used."
+    },
+    { "defnode", 2, 3, cmd_defnode, "defnode <NAME> <EXPR> [<VALUE>]",
+      "Define the variable NAME to the result of evalutating EXPR, which\n"
+      "        must be a nodeset. If no node matching EXPR exists yet, one\n"
+      "        is created and NAME will refer to it. If VALUE is given, this\n"
+      "        is the same as 'set EXPR VALUE'; if VALUE is not given, the\n"
+      "        node is created as if with 'clear EXPR' would and NAME refers\n"
+      "        to that node."
+    },
     { "help", 0, 0, cmd_help, "help",
       "Print this help text"
     },
@@ -495,6 +558,7 @@ static void usage(void) {
     fprintf(stderr, "  -r, --root ROOT    use ROOT as the root of the filesystem\n");
     fprintf(stderr, "  -I, --include DIR  search DIR for modules; can be given mutiple times\n");
     fprintf(stderr, "  --nostdinc         do not search the builtin default directories for modules\n");
+    fprintf(stderr, "  --noload           do not load any files into the tree on startup\n");
 
     exit(EXIT_FAILURE);
 }
@@ -503,7 +567,8 @@ static void parse_opts(int argc, char **argv) {
     int opt;
     size_t loadpathlen = 0;
     enum {
-        VAL_NO_STDINC = CHAR_MAX + 1
+        VAL_NO_STDINC = CHAR_MAX + 1,
+        VAL_NO_LOAD = VAL_NO_STDINC + 1
     };
     struct option options[] = {
         { "help",      0, 0, 'h' },
@@ -513,6 +578,7 @@ static void parse_opts(int argc, char **argv) {
         { "root",      1, 0, 'r' },
         { "include",   1, 0, 'I' },
         { "nostdinc",  0, 0, VAL_NO_STDINC },
+        { "noload",    0, 0, VAL_NO_LOAD },
         { 0, 0, 0, 0}
     };
     int idx;
@@ -540,6 +606,9 @@ static void parse_opts(int argc, char **argv) {
         case VAL_NO_STDINC:
             flags |= AUG_NO_STDINC;
             break;
+        case VAL_NO_LOAD:
+            flags |= AUG_NO_LOAD;
+            break;
         default:
             usage();
             break;
