@@ -163,7 +163,9 @@ char *unescape(const char *s, int len) {
             i += 1;
         }
 
-    CALLOC(result, size + 1);
+    if (ALLOC_N(result, size + 1) < 0)
+        return NULL;
+
     for (i = 0, t = result; i < len; i++, size++) {
         if (s[i] == '\\' && (n = strchr(escape_names, s[i+1])) != NULL) {
             *t++ = escape_chars[n - escape_names];
@@ -191,7 +193,8 @@ char *escape(const char *text, int cnt) {
         else
             len += 1;
     }
-    CALLOC(esc, len+1);
+    if (ALLOC_N(esc, len+1) < 0)
+        return NULL;
     e = esc;
     for (int i=0; i < cnt; i++) {
         char *p;
@@ -302,7 +305,77 @@ int close_memstream(struct memstream *ms) {
     return 0;
 }
 
+char *path_expand(struct tree *tree, const char *ppath) {
+    struct tree *siblings = tree->parent->children;
 
+    char *path;
+    const char *label;
+    int cnt = 0, ind = 0, r;
+
+    list_for_each(t, siblings) {
+        if (streqv(t->label, tree->label)) {
+            cnt += 1;
+            if (t == tree)
+                ind = cnt;
+        }
+    }
+
+    if (ppath == NULL)
+        ppath = "";
+
+    if (tree == NULL)
+        label = "(no_tree)";
+    else if (tree->label == NULL)
+        label = "(none)";
+    else
+        label = tree->label;
+
+    if (cnt > 1) {
+        r = asprintf(&path, "%s/%s[%d]", ppath, label, ind);
+    } else {
+        r = asprintf(&path, "%s/%s", ppath, label);
+    }
+    if (r == -1)
+        return NULL;
+    return path;
+}
+
+char *path_of_tree(struct tree *tree) {
+    int depth, i;
+    struct tree *t, **anc;
+    char *path = NULL;
+
+    for (t = tree, depth = 1; ! ROOT_P(t); depth++, t = t->parent);
+    if (ALLOC_N(anc, depth) < 0)
+        return NULL;
+
+    for (t = tree, i = depth - 1; i >= 0; i--, t = t->parent)
+        anc[i] = t;
+
+    for (i = 0; i < depth; i++) {
+        char *p = path_expand(anc[i], path);
+        free(path);
+        path = p;
+    }
+    FREE(anc);
+    return path;
+}
+
+const char *xstrerror(int errnum, char *buf, size_t len) {
+#ifdef HAVE_STRERROR_R
+# ifdef __USE_GNU
+    /* Annoying linux specific API contract */
+    return strerror_r(errnum, buf, len);
+# else
+    strerror_r(errnum, buf, len);
+    return buf;
+# endif
+#else
+    int n = snprintf(buf, len, "errno=%d", errnum);
+    return (0 < n && n < len
+            ? buf : "internal error: buffer too small in xstrerror");
+#endif
+}
 
 /*
  * Local variables:
