@@ -23,8 +23,6 @@
 #ifndef INTERNAL_H_
 #define INTERNAL_H_
 
-#define DEBUG
-
 #include "list.h"
 #include "datadir.h"
 
@@ -32,6 +30,7 @@
 #include <string.h>
 #include <strings.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <errno.h>
 #include <assert.h>
@@ -44,6 +43,9 @@
  * The default location for lens definitions */
 #define AUGEAS_LENS_DIR DATADIR "/augeas/lenses"
 
+/* The directory where we install lenses distribute with Augeas */
+#define AUGEAS_LENS_DIST_DIR DATADIR "/augeas/lenses/dist"
+
 /* Define: AUGEAS_ROOT_ENV
  * The env var that points to the chroot holding files we may modify.
  * Mostly useful for testing */
@@ -83,6 +85,9 @@
 
 #define AUGEAS_EVENTS_SAVED AUGEAS_EVENTS "/saved"
 
+/* Where to put information about parsing of path expressions */
+#define AUGEAS_META_PATHX AUGEAS_META_TREE "/pathx"
+
 /* Define: AUGEAS_LENS_ENV
  * Name of env var that contains list of paths to search for additional
    spec files */
@@ -97,6 +102,12 @@
  * Character separating paths in a list of paths */
 #define PATH_SEP_CHAR ':'
 
+/* Constants for setting the save mode via the augeas path at
+ * AUGEAS_META_SAVE_MODE */
+#define AUG_SAVE_BACKUP_TEXT "backup"
+#define AUG_SAVE_NEWFILE_TEXT "newfile"
+#define AUG_SAVE_NOOP_TEXT "noop"
+#define AUG_SAVE_OVERWRITE_TEXT "overwrite"
 
 #ifdef __GNUC__
 
@@ -142,6 +153,8 @@
 #define ATTRIBUTE_RETURN_CHECK
 #endif                                   /* __GNUC__ */
 
+#define ARRAY_CARDINALITY(array) (sizeof (array) / sizeof *(array))
+
 /* String equality tests, suggested by Jim Meyering. */
 #define STREQ(a,b) (strcmp((a),(b)) == 0)
 #define STRCASEEQ(a,b) (strcasecmp((a),(b)) == 0)
@@ -245,10 +258,14 @@ char *format_pos(const char *text, int pos);
  */
 char* read_file(const char *path);
 
-/* Define: AUG_NO_DEFAULT_LOAD
- * A hidden flag used by augparse to suppress loading of all the modules
+/* Get the error message for ERRNUM in a threadsafe way. Based on libvirt's
+ * virStrError
+ */
+const char *xstrerror(int errnum, char *buf, size_t len);
+
+/* A hidden flag used by augparse to suppress loading of all the modules
  * on the path */
-#define AUG_NO_DEFAULT_LOAD (1 << 15)
+#define AUG_NO_MODL_AUTOLOAD (1 << 15)
 
 /* Struct: augeas
  * The data structure representing a connection to Augeas. */
@@ -261,6 +278,7 @@ struct augeas {
     size_t            nmodpath;
     char             *modpathz;   /* The search path for modules as a
                                      glibc argz vector */
+    struct pathx_symtab *symtab;
 };
 
 /* Struct: tree
@@ -271,6 +289,12 @@ struct augeas {
  * standalone trees with a fake root, called origin. That root is generally
  * not referenced from anywhere. Standalone trees should be created with
  * MAKE_TREE_ORIGIN.
+ *
+ * The DIRTY flag is used to track which parts of the tree might need to be
+ * saved. For any node that is marked dirty, all of its ancestors must be
+ * marked dirty, too. Instead of setting this flag directly, the function
+ * TREE_MARK_DIRTY in augeas.c should be used (and only functions in that
+ * file should have a need to mark nodes as dirty)
  */
 struct tree {
     struct tree *next;
@@ -281,6 +305,11 @@ struct tree {
     int          dirty;
 };
 
+/* The opaque structure used to represent path expressions. API's
+ * using STRUCT PATHX are declared farther below
+ */
+struct pathx;
+
 #define ROOT_P(t) ((t) != NULL && (t)->parent == (t)->parent->parent)
 
 /* Function: make_tree
@@ -296,16 +325,19 @@ struct tree *make_tree(char *label, char *value,
  */
 struct tree  *make_tree_origin(struct tree *root);
 
-int aug_tree_replace(struct augeas *aug, const char *path, struct tree *sub);
+int tree_replace(struct tree *origin, const char *path, struct tree *sub);
+/* Make a new tree node and append it to parent's children */
+struct tree *tree_append(struct tree *parent, char *label, char *value);
 
-int tree_rm(struct tree *origin, const char *path);
-struct tree *tree_set(struct tree *tree, const char *path, const char *value);
-int tree_insert(struct tree *origin, const char *path, const char *label,
-                int before);
+int tree_rm(struct pathx *p);
+int tree_unlink(struct tree *tree);
+struct tree *tree_set(struct pathx *p, const char *value);
+int tree_insert(struct pathx *p, const char *label, int before);
 int free_tree(struct tree *tree);
-int print_tree(const struct tree *tree, FILE *out, const char *path,
-               int pr_hidden);
+int dump_tree(FILE *out, struct tree *tree);
 int tree_equal(const struct tree *t1, const struct tree *t2);
+char *path_expand(struct tree *tree, const char *ppath);
+char *path_of_tree(struct tree *tree);
 
 /* Struct: memstream
  * Wrappers to simulate OPEN_MEMSTREAM where that's not available. The
@@ -342,15 +374,54 @@ int close_memstream(struct memstream *ms);
  * Path expressions
  */
 
-struct path;
+typedef enum {
+    PATHX_NOERROR = 0,
+    PATHX_ENAME,
+    PATHX_ESTRING,
+    PATHX_ENUMBER,
+    PATHX_EDELIM,
+    PATHX_ENOEQUAL,
+    PATHX_ENOMEM,
+    PATHX_EPRED,
+    PATHX_EPAREN,
+    PATHX_ESLASH,
+    PATHX_EINTERNAL,
+    PATHX_ETYPE,
+    PATHX_ENOVAR,
+    PATHX_EEND,
+    PATHX_ENONODES
+} pathx_errcode_t;
 
-struct path *make_path(const struct tree *root, const char *path);
-struct tree *path_first(struct path *path);
-struct tree *path_next(struct path *path, struct tree *cur);
-int path_find_one(struct path *path, struct tree **match);
-int path_expand_tree(struct path *path, struct tree **tree);
-void free_path(struct path *path);
+struct pathx;
+struct pathx_symtab;
 
+const char *pathx_error(struct pathx *pathx, const char **txt, int *pos);
+
+/* Parse the string PATH into a path expression PX that will be evaluated
+ * against the tree ORIGIN.
+ *
+ * If NEED_NODESET is true, the resulting path expression must evaluate toa
+ * nodeset, otherwise it can evaluate to a value of any type.
+ *
+ * Returns 0 on success, and -1 on error
+ */
+int pathx_parse(const struct tree *origin, const char *path,
+                bool need_nodeset,
+                struct pathx_symtab *symtab,
+                struct pathx **px);
+struct tree *pathx_first(struct pathx *path);
+struct tree *pathx_next(struct pathx *path);
+int pathx_find_one(struct pathx *path, struct tree **match);
+int pathx_expand_tree(struct pathx *path, struct tree **tree);
+void free_pathx(struct pathx *path);
+
+int pathx_symtab_init(struct pathx_symtab **symtab);
+int pathx_symtab_define(struct pathx_symtab **symtab,
+                        const char *name, struct pathx *px);
+int pathx_symtab_undefine(struct pathx_symtab **symtab, const char *name);
+void pathx_symtab_remove_descendants(struct pathx *pathx,
+                                     const struct tree *tree);
+void free_symtab(struct pathx_symtab *symtab);
 #endif
 
 
