@@ -41,6 +41,44 @@ static const char *const tags[] = {
     "subtree", "star", "maybe"
 };
 
+/* Construct a finite automaton from REGEXP and return it in *FA.
+ *
+ * Return NULL if REGEXP is valid, if the regexp REGEXP has syntax errors,
+ * return an exception.
+ */
+static struct value *str_to_fa(struct info *info, const char *pattern,
+                               struct fa **fa) {
+    int error;
+    struct value *exn = NULL;
+    size_t re_err_len;
+    char *re_str, *re_err;
+
+    error = fa_compile(pattern, strlen(pattern), fa);
+    if (error == REG_NOERROR)
+        return NULL;
+
+    re_str = escape(pattern, -1);
+    if (re_str == NULL) {
+        FIXME("Out of memory");
+    }
+    exn = make_exn_value(info, "Invalid regular expression /%s/", re_str);
+
+    re_err_len = regerror(error, NULL, NULL, 0);
+    if (ALLOC_N(re_err, re_err_len) < 0) {
+        FIXME("Out of memory");
+    }
+    regerror(error, NULL, re_err, re_err_len);
+    exn_printf_line(exn, "%s", re_err);
+
+    free(re_str);
+    free(re_err);
+    return exn;
+}
+
+static struct value *regexp_to_fa(struct regexp *regexp, struct fa **fa) {
+    return str_to_fa(regexp->info, regexp->pattern->str, fa);
+}
+
 static struct lens *make_lens(enum lens_tag tag, struct info *info) {
     struct lens *lens;
     make_ref(lens);
@@ -243,23 +281,20 @@ struct value *lns_make_prim(enum lens_tag tag, struct info *info,
                             struct regexp *regexp, struct string *string) {
     struct lens *lens = NULL;
     struct value *exn = NULL;
-    fa_t fa_slash = NULL;
-    fa_t fa_key = NULL;
-    fa_t fa_isect = NULL;
+    struct fa *fa_slash = NULL;
+    struct fa *fa_key = NULL;
+    struct fa *fa_isect = NULL;
 
     /* Typecheck */
     if (tag == L_KEY) {
-        int error = fa_compile("(.|\n)*/(.|\n)*", &fa_slash);
-        if (error != REG_NOERROR) {
-            exn = make_exn_value(info,
-                                 "unexpected error from fa_compile %d", error);
+        exn = str_to_fa(info, "(.|\n)*/(.|\n)*", &fa_slash);
+        if (exn != NULL)
             goto error;
-        }
-        fa_key = regexp_to_fa(regexp);
-        if (fa_key == NULL) {
-            exn = make_exn_value(info, "fa_compile of key failed");
+
+        exn = regexp_to_fa(regexp, &fa_key);
+        if (exn != NULL)
             goto error;
-        }
+
         fa_isect = fa_intersect(fa_slash, fa_key);
         if (! fa_is_basic(fa_isect, FA_EMPTY)) {
             exn = make_exn_value(info,
@@ -322,21 +357,24 @@ struct value *lns_make_prim(enum lens_tag tag, struct info *info,
  */
 static struct value *disjoint_check(struct info *info, const char *msg,
                                     struct regexp *r1, struct regexp *r2) {
-    fa_t fa1 = regexp_to_fa(r1);
-    fa_t fa2 = regexp_to_fa(r2);
-    fa_t fa = NULL;
+    struct fa *fa1 = NULL;
+    struct fa *fa2 = NULL;
+    struct fa *fa = NULL;
     struct value *exn = NULL;
 
-    if (fa1 == NULL || fa2 == NULL) {
-        fa_free(fa1);
-        fa_free(fa2);
-        return make_exn_value(ref(info),
-              "internal error: compile in disjoint_check failed");
-    }
+    exn = regexp_to_fa(r1, &fa1);
+    if (exn != NULL)
+        goto done;
+
+    exn = regexp_to_fa(r2, &fa2);
+    if (exn != NULL)
+        goto done;
 
     fa = fa_intersect(fa1, fa2);
     if (! fa_is_basic(fa, FA_EMPTY)) {
-        char *xmpl = fa_example(fa);
+        size_t xmpl_len;
+        char *xmpl;
+        fa_example(fa, &xmpl, &xmpl_len);
         exn = make_exn_value(ref(info),
                              "overlapping lenses in %s", msg);
 
@@ -344,6 +382,7 @@ static struct value *disjoint_check(struct info *info, const char *msg,
         free(xmpl);
     }
 
+ done:
     fa_free(fa);
     fa_free(fa1);
     fa_free(fa2);
@@ -371,10 +410,11 @@ static struct value *typecheck_union(struct info *info,
     return exn;
 }
 
-static struct value *ambig_check(struct info *info, fa_t fa1, fa_t fa2,
+static struct value *ambig_check(struct info *info, struct fa *fa1, struct fa *fa2,
                                  const char *msg) {
     char *upv, *pv, *v;
-    upv = fa_ambig_example(fa1, fa2, &pv, &v);
+    size_t upv_len;
+    fa_ambig_example(fa1, fa2, &upv, &upv_len, &pv, &v);
     struct value *exn = NULL;
 
     if (upv != NULL) {
@@ -400,17 +440,20 @@ static struct value *ambig_check(struct info *info, fa_t fa1, fa_t fa2,
 
 static struct value *ambig_concat_check(struct info *info, const char *msg,
                                         struct regexp *r1, struct regexp *r2) {
-    fa_t fa1 = regexp_to_fa(r1);
-    fa_t fa2 = regexp_to_fa(r2);
+    struct fa *fa1 = NULL;
+    struct fa *fa2 = NULL;
     struct value *result = NULL;
 
-    if (fa1 == NULL || fa2 == NULL) {
-        fa_free(fa1);
-        fa_free(fa2);
-        return make_exn_value(ref(info), "Internal error: regexp_to_fa failed");
-    }
+    result = regexp_to_fa(r1, &fa1);
+    if (result != NULL)
+        goto done;
+
+    result = regexp_to_fa(r2, &fa2);
+    if (result != NULL)
+        goto done;
 
     result = ambig_check(info, fa1, fa2, msg);
+ done:
     fa_free(fa1);
     fa_free(fa2);
     return result;
@@ -439,14 +482,18 @@ static struct value *typecheck_concat(struct info *info,
 
 static struct value *ambig_iter_check(struct info *info, const char *msg,
                                       struct regexp *r) {
-    fa_t fas, fa;
+    struct fa *fas = NULL, *fa = NULL;
     struct value *result = NULL;
 
-    fa = regexp_to_fa(r);
+    result = regexp_to_fa(r, &fa);
+    if (result != NULL)
+        goto done;
+
     fas = fa_iter(fa, 0, -1);
 
     result = ambig_check(info, fa, fas, msg);
 
+ done:
     fa_free(fa);
     fa_free(fas);
     return result;
@@ -582,13 +629,15 @@ static struct regexp *lns_key_regexp(struct lens *l, struct value **exn) {
                 struct regexp *r = lns_key_regexp(l->children[i], exn);
                 if (*exn != NULL)
                     return NULL;
-                if (k == NULL) {
-                    k = r;
-                } else {
-                    struct regexp *u = regexp_union(l->info, k, r);
-                    unref(k, regexp);
-                    unref(r, regexp);
-                    k = u;
+                if (r != NULL) {
+                    if (k == NULL) {
+                        k = r;
+                    } else {
+                        struct regexp *u = regexp_union(l->info, k, r);
+                        unref(k, regexp);
+                        unref(r, regexp);
+                        k = u;
+                    }
                 }
             }
             return k;
@@ -647,6 +696,8 @@ void free_lens(struct lens *lens) {
 }
 
 void lens_release(struct lens *lens) {
+    if (lens == NULL)
+        return;
     regexp_release(lens->ctype);
     regexp_release(lens->atype);
     if (lens->tag == L_KEY || lens->tag == L_STORE)
