@@ -25,8 +25,6 @@
 #include <stdarg.h>
 #include "syntax.h"
 
-//#define DEBUG_SPEW
-
 /* Data structure to keep track of where we are in the tree. The split
  * describes a sublist of the list of siblings in the current tree. The
  * put_* functions don't operate on the tree directly, instead they operate
@@ -65,22 +63,6 @@ struct state {
 static void create_lens(struct lens *lens, struct state *state);
 static void put_lens(struct lens *lens, struct state *state);
 
-#ifdef DEBUG_SPEW
-#define _l(obj) (((obj) == NULL) ? "nil" : (obj)->label)
-
-static const char *_t(struct lens *lens) {
-    static const char *types[] = {
-        "action", "[]", "literal", "name", "|", ".",
-        "rule", "abbrev", "+", "*", "?"
-    };
-    return types[match->type];
-}
-
-#define debug(fmt, args ...) printf(fmt, ## args)
-#else
-#define debug(fmt, args ...)
-#endif
-
 static void put_error(struct state *state, struct lens *lens,
                       const char *format, ...)
 {
@@ -106,6 +88,26 @@ static void put_error(struct state *state, struct lens *lens,
         state->error->message = NULL;
 }
 
+static void regexp_match_error(struct state *state, struct lens *lens,
+                               int count, struct split *split,
+                               struct regexp *r) {
+    char *text = strndup(split->labels + split->start,
+                         split->end - split->start);
+    char *pat = regexp_escape(r);
+
+    if (count == -1) {
+        put_error(state, lens, "Failed to match /%s/ with %s", pat, text);
+    } else if (count == -2) {
+        put_error(state, lens, "Internal error matching /%s/ with %s",
+                  pat, text);
+    } else if (count == -3) {
+        /* Should have been caught by the typechecker */
+        put_error(state, lens, "Syntax error in regexp /%s/", pat);
+    }
+    free(pat);
+    free(text);
+}
+
 static struct split *make_split(struct tree *tree) {
     struct split *split;
     CALLOC(split, 1);
@@ -187,20 +189,13 @@ static struct split *split_concat(struct state *state, struct lens *lens) {
         return split;
     }
 
-    if (atype->re != NULL)
-        atype->re->regs_allocated = REGS_UNALLOCATED;
+    MEMZERO(&regs, 1);
     count = regexp_match(atype, outer->labels, outer->end,
                          outer->start, &regs);
-    if (count == -2) {
-        FIXME("Match failed - produce better error");
-        abort();
-    } else if (count == -1 || count != outer->end - outer->start) {
-        char *labels = strndup(outer->labels + outer->start,
-                               outer->end - outer->start);
-        put_error(state, lens,
-                  "Failed to match /%s/ with %s",
-                  atype->pattern->str, labels);
-        free(labels);
+    if (count >= 0 && count != outer->end - outer->start)
+        count = -1;
+    if (count < 0) {
+        regexp_match_error(state, lens, count, outer, atype);
         return NULL;
     }
 
@@ -225,10 +220,11 @@ static struct split *split_concat(struct state *state, struct lens *lens) {
     return split;
 }
 
-static struct split *split_iter(struct lens *lens, struct split *outer) {
+static struct split *split_iter(struct state *state, struct lens *lens) {
     assert(lens->tag == L_STAR);
 
     int count = 0;
+    struct split *outer = state->split;
     struct split *split = NULL;
     struct regexp *atype = lens->child->atype;
 
@@ -237,12 +233,13 @@ static struct split *split_iter(struct lens *lens, struct split *outer) {
     struct split *tail = NULL;
     while (pos < outer->end) {
         count = regexp_match(atype, outer->labels, outer->end, pos, NULL);
-        if (count == -2) {
-            FIXME("Match failed - produce better error");
-            abort();
-        } else if (count == -1) {
+        if (count == -1) {
             break;
+        } else if (count < -1) {
+            regexp_match_error(state, lens, count, outer, atype);
+            goto error;
         }
+
         struct tree *follow = cur;
         for (int j = pos; j < pos + count; j++) {
             if (outer->labels[j] == '/')
@@ -254,6 +251,9 @@ static struct split *split_iter(struct lens *lens, struct split *outer) {
         pos += count;
     }
     return split;
+ error:
+    free_split(split);
+    return NULL;
 }
 
 /* Check if LENS applies to the current split in STATE */
@@ -263,10 +263,11 @@ static int applies(struct lens *lens, struct state *state) {
 
     count = regexp_match(lens->atype, split->labels, split->end,
                          split->start, NULL);
-    if (count == -2) {
-        FIXME("Match failed - produce better error");
-        abort();
+    if (count < -1) {
+        regexp_match_error(state, lens, count, split, lens->atype);
+        return 0;
     }
+
     if (count != split->end - split->start)
         return 0;
     if (count == 0 && lens->value)
@@ -475,7 +476,7 @@ static void put_quant_star(struct lens *lens, struct state *state) {
     struct split *oldsplit = state->split;
     struct skel *oldskel = state->skel;
 
-    struct split *split = split_iter(lens, state->split);
+    struct split *split = split_iter(state, lens);
 
     state->skel = state->skel->skels;
     set_split(state, split);
@@ -523,8 +524,6 @@ static void put_store(struct lens *lens, struct state *state) {
 }
 
 static void put_lens(struct lens *lens, struct state *state) {
-    debug("put_lens: %s:%d %s\n", _t(lens), lens->info->first_line,
-          _l(state->tree));
     if (state->error != NULL)
         return;
 
@@ -599,7 +598,8 @@ static void create_concat(struct lens *lens, struct state *state) {
     set_split(state, split);
     for (int i=0; i < lens->nchildren; i++) {
         if (state->split == NULL) {
-            syntax_error(lens->info, "Short split for concat");
+            put_error(state, lens,
+                      "Not enough components in concat");
             list_free(split);
             return;
         }
@@ -614,7 +614,7 @@ static void create_quant_star(struct lens *lens, struct state *state) {
     assert(lens->tag == L_STAR);
     struct split *oldsplit = state->split;
 
-    struct split *split = split_iter(lens, state->split);
+    struct split *split = split_iter(state, lens);
 
     set_split(state, split);
     while (state->split != NULL) {
@@ -637,8 +637,6 @@ static void create_quant_maybe(struct lens *lens, struct state *state) {
 }
 
 static void create_lens(struct lens *lens, struct state *state) {
-    debug("create_lens: %s:%d %s\n", _t(lens), lens->info->first_line,
-          _l(state->tree));
     if (state->error != NULL)
         return;
     switch(lens->tag) {
@@ -718,35 +716,6 @@ void lns_put(FILE *out, struct lens *lens, struct tree *tree,
     }
 }
 
-/*
-  Cleverly created with Elisp:
-(defun my-ins-put ()
-  (interactive)
-  (insert "// BEGIN\n")
-  (let
-      ((types '("ACTION" "SUBTREE" "LITERAL" "NAME" "ALTERNATIVE"
-                "SEQUENCE" "RULE_REF" "ABBREV_REF"
-                "QUANT_PLUS" "QUANT_STAR" "QUANT_MAYBE")))
-    (mapcar
-     (lambda (type)
-       (insert "static void put_" (downcase type) "(struct match *match, struct state *state) {
-    assert(match->type == " type ");
-}
-
-"))
-          types)
-    (mapcar
-     (lambda (type)
-       (insert "static void create_" (downcase type) "(struct match *match, struct state *state) {
-    assert(match->type == " type ");
-
-}
-
-"))
-          types)
-  (insert "// END\n")))
- */
-
 /*
  * Local variables:
  *  indent-tabs-mode: nil
