@@ -31,6 +31,26 @@ static const struct string empty_pattern_string = {
 
 static const struct string *const empty_pattern = &empty_pattern_string;
 
+char *regexp_escape(struct regexp *r) {
+    char *pat = escape(r->pattern->str, -1);
+
+    if (pat == NULL)
+        return NULL;
+
+    /* Remove unneeded '()' from pat */
+    for (int changed = 1; changed;) {
+        changed = 0;
+        for (char *p = pat; *p != '\0'; p++) {
+            if (*p == '(' && p[1] == ')') {
+                memmove(p, p+2, strlen(p+2)+1);
+                changed = 1;
+            }
+        }
+    }
+
+    return pat;
+}
+
 void print_regexp(FILE *out, struct regexp *r) {
     if (r == NULL) {
         fprintf(out, "<NULL>");
@@ -167,15 +187,16 @@ regexp_minus(struct info *info, struct regexp *r1, struct regexp *r2) {
     const char *p1 = r1->pattern->str;
     const char *p2 = r2->pattern->str;
     struct regexp *result = NULL;
-    fa_t fa = NULL, fa1 = NULL, fa2 = NULL;
+    struct fa *fa = NULL, *fa1 = NULL, *fa2 = NULL;
     int r;
     char *s = NULL;
+    size_t s_len;
 
-    r = fa_compile(p1, &fa1);
+    r = fa_compile(p1, strlen(p1), &fa1);
     if (r != REG_NOERROR)
         goto error;
 
-    r = fa_compile(p2, &fa2);
+    r = fa_compile(p2, strlen(p2), &fa2);
     if (r != REG_NOERROR)
         goto error;
 
@@ -183,7 +204,7 @@ regexp_minus(struct info *info, struct regexp *r1, struct regexp *r2) {
     if (fa == NULL)
         goto error;
 
-    r = fa_as_regexp(fa, &s);
+    r = fa_as_regexp(fa, &s, &s_len);
     if (r < 0)
         goto error;
 
@@ -209,10 +230,14 @@ regexp_minus(struct info *info, struct regexp *r1, struct regexp *r2) {
 
 struct regexp *
 regexp_iter(struct info *info, struct regexp *r, int min, int max) {
-    const char *p = r->pattern->str;
+    const char *p;
     char *s;
     int ret = 0;
 
+    if (r == NULL)
+        return NULL;
+
+    p = r->pattern->str;
     if ((min == 0 || min == 1) && max == -1) {
         char q = (min == 0) ? '*' : '+';
         ret = asprintf(&s, "(%s)%c", p, q);
@@ -226,10 +251,13 @@ regexp_iter(struct info *info, struct regexp *r, int min, int max) {
 
 struct regexp *
 regexp_maybe(struct info *info, struct regexp *r) {
-    const char *p = r->pattern->str;
+    const char *p;
     char *s;
     int ret;
 
+    if (r == NULL)
+        return NULL;
+    p = r->pattern->str;
     ret = asprintf(&s, "(%s)?", p);
     return (ret == -1) ? NULL : make_regexp(info, s);
 }
@@ -248,13 +276,27 @@ struct regexp *regexp_make_empty(struct info *info) {
 }
 
 int regexp_compile(struct regexp *r) {
+    /* See the GNU regex manual or regex.h in gnulib for
+     * an explanation of these flags. They are set so that the regex
+     * matcher interprets regular expressions the same way that libfa
+     * does
+     */
+    static const reg_syntax_t syntax =
+        RE_CONTEXT_INDEP_OPS|RE_CONTEXT_INVALID_OPS|RE_DOT_NOT_NULL
+        |RE_INTERVALS|RE_NO_BK_BRACES|RE_NO_BK_PARENS|RE_NO_BK_REFS
+        |RE_NO_BK_VBAR|RE_NO_EMPTY_RANGES
+        |RE_NO_POSIX_BACKTRACKING|RE_CONTEXT_INVALID_DUP|RE_NO_GNU_OPS;
+    reg_syntax_t old_syntax = re_syntax_options;
+    const char *c = NULL;
+
     if (r->re == NULL)
         CALLOC(r->re, 1);
 
-    re_syntax_options = RE_SYNTAX_POSIX_MINIMAL_EXTENDED & ~(RE_DOT_NEWLINE);
-    const char *c =
-        re_compile_pattern(r->pattern->str, strlen(r->pattern->str), r->re);
+    re_syntax_options = syntax;
+    c = re_compile_pattern(r->pattern->str, strlen(r->pattern->str), r->re);
+    re_syntax_options = old_syntax;
 
+    r->re->regs_allocated = REGS_REALLOCATE;
     if (c != NULL) {
         char *p = escape(r->pattern->str, -1);
         syntax_error(r->info, "invalid regexp /%s/: %s", p, c);
@@ -285,18 +327,6 @@ int regexp_nsub(struct regexp *r) {
     return r->re->re_nsub;
 }
 
-fa_t regexp_to_fa(struct regexp *regexp) {
-    fa_t fa;
-    int error = fa_compile(regexp->pattern->str, &fa);
-    if (error != REG_NOERROR) {
-        syntax_error(regexp->info,
-                     "unexpected error from fa_compile %d compiling %s",
-                     error, regexp->pattern->str);
-        return NULL;
-    }
-    return fa;
-}
-
 void regexp_release(struct regexp *regexp) {
     if (regexp->re != NULL) {
         regfree(regexp->re);
