@@ -35,6 +35,7 @@
 #include "memory.h"
 #include "syntax.h"
 #include "augeas.h"
+#include "transform.h"
 
 /* Extension of source files */
 #define AUG_EXT ".aug"
@@ -358,6 +359,18 @@ struct string *make_string(char *str) {
     return string;
 }
 
+struct string *dup_string(const char *str) {
+    struct string *string;
+    make_ref(string);
+    if (str == NULL)
+        string->str = strdup("");
+    else
+        string->str = strdup(str);
+    if (string->str == NULL)
+        unref(string, string);
+    return string;
+}
+
 struct term *make_app_term(struct term *lambda, struct term *arg,
                            struct info *info) {
   struct term *app = make_term(A_APP, info);
@@ -470,6 +483,59 @@ static struct binding *bnd_lookup(struct binding *bindings, const char *name) {
     return NULL;
 }
 
+static char *modname_of_qname(const char *qname) {
+    char *dot = strchr(qname, '.');
+    if (dot == NULL)
+        return NULL;
+
+    return strndup(qname, dot - qname);
+}
+
+static int lookup_internal(struct augeas *aug, const char *ctx_modname,
+                           const char *name, struct binding **bnd) {
+    char *modname = modname_of_qname(name);
+
+    *bnd = NULL;
+
+    if (modname == NULL) {
+        struct module *builtin =
+            module_find(aug->modules, builtin_module);
+        assert(builtin != NULL);
+        *bnd = bnd_lookup(builtin->bindings, name);
+        return 0;
+    }
+
+ qual_lookup:
+    list_for_each(module, aug->modules) {
+        if (STRCASEEQ(module->name, modname)) {
+            *bnd = bnd_lookup(module->bindings, name + strlen(modname) + 1);
+            free(modname);
+            return 0;
+        }
+    }
+    /* Try to load the module */
+    if (streqv(modname, ctx_modname)) {
+        free(modname);
+        return 0;
+    }
+    int loaded = load_module(aug, modname) == 0;
+    if (loaded)
+        goto qual_lookup;
+
+    free(modname);
+    return -1;
+}
+
+struct lens *lens_lookup(struct augeas *aug, const char *qname) {
+    struct binding *bnd = NULL;
+
+    if (lookup_internal(aug, NULL, qname, &bnd) < 0)
+        return NULL;
+    if (bnd == NULL || bnd->value->tag != V_LENS)
+        return NULL;
+    return bnd->value->lens;
+}
+
 static struct binding *ctx_lookup_bnd(struct info *info,
                                       struct ctx *ctx, const char *name) {
     struct binding *b = NULL;
@@ -483,35 +549,15 @@ static struct binding *ctx_lookup_bnd(struct info *info,
         return b;
 
     if (ctx->aug != NULL) {
-        char *dot = strchr(name, '.');
-        if (dot != NULL) {
-        qual_lookup:
-            list_for_each(module, ctx->aug->modules) {
-                if (STRCASEEQLEN(module->name, name, strlen(module->name))
-                    && dot - name == strlen(module->name))
-                    return bnd_lookup(module->bindings, dot + 1);
-            }
-            /* Try to load the module */
-            char *modname = strndup(name, dot - name);
-            if (STREQ(modname, ctx->name)) {
-                free(modname);
-                return NULL;
-            }
-            int loaded = load_module(ctx->aug, modname) == 0;
-            if (loaded) {
-                free(modname);
-                goto qual_lookup;
-            } else {
-                syntax_error(info, "Could not load module %s for %s",
-                             modname, name);
-                free(modname);
-            }
-        } else {
-            struct module *builtin =
-                module_find(ctx->aug->modules, builtin_module);
-            assert(builtin != NULL);
-            return bnd_lookup(builtin->bindings, name);
-        }
+        int r;
+        r = lookup_internal(ctx->aug, ctx->name, name, &b);
+        if (r == 0)
+            return b;
+        char *modname = modname_of_qname(name);
+        syntax_error(info, "Could not load module %s for %s",
+                     modname, name);
+        free(modname);
+        return NULL;
     }
     return NULL;
 }
@@ -625,7 +671,7 @@ static void print_value(FILE *out, struct value *v) {
         fprintf(out, ">");
         break;
     case V_TREE:
-        print_tree(v->origin->children, stdout, "/*" , 1);
+        dump_tree(stdout, v->origin);
         break;
     case V_FILTER:
         fprintf(out, "<filter:");
@@ -1596,6 +1642,25 @@ static struct value *compile_exp(struct info *info,
     return v;
 }
 
+static void print_tree(FILE *out, int indent, struct tree *tree) {
+    list_for_each(t, tree) {
+        for (int i=0; i < indent; i++) fputc(' ', out);
+        fprintf(out, "{ ");
+        if (t->label != NULL)
+            fprintf(out, "\"%s\"", t->label);
+        if (t->value != NULL)
+            fprintf(out, " = \"%s\"", t->value);
+        if (t->children != NULL) {
+            fputc('\n', out);
+            print_tree(out, indent + 2, t->children);
+            for (int i=0; i < indent; i++) fputc(' ', out);
+        } else {
+            fputc(' ', out);
+        }
+        fprintf(out, "}\n");
+    }
+}
+
 static int compile_test(struct term *term, struct ctx *ctx) {
     struct value *actual = compile_exp(term->info, term->test, ctx);
     struct value *expect = NULL;
@@ -1634,7 +1699,11 @@ static int compile_test(struct term *term, struct ctx *ctx) {
             printf("Test result: ");
             print_info(stdout, term->info);
             printf("\n");
-            print_value(stdout, actual);
+            if (actual->tag == V_TREE) {
+                print_tree(stdout, 2, actual->origin->children);
+            } else {
+                print_value(stdout, actual);
+            }
             printf("\n");
         }
     }
@@ -1861,7 +1930,7 @@ int interpreter_init(struct augeas *aug) {
 
     aug->modules = builtin_init();
 
-    if (aug->flags & AUG_NO_DEFAULT_LOAD)
+    if (aug->flags & AUG_NO_MODL_AUTOLOAD)
         return 0;
 
     // For now, we just load every file on the search path
