@@ -31,13 +31,15 @@
 #include <sys/stat.h>
 #include <unistd.h>
 #include <selinux/selinux.h>
+#include <stdbool.h>
 
 #include "internal.h"
 #include "memory.h"
 #include "augeas.h"
 #include "syntax.h"
+#include "transform.h"
 
-static const int fnm_flags = FNM_FILE_NAME;
+static const int fnm_flags = FNM_PATHNAME;
 static const int glob_flags = GLOB_NOSORT;
 
 /* Extension for newly created files */
@@ -56,15 +58,17 @@ static const int glob_flags = GLOB_NOSORT;
  *   error/path: path to tree node where error occurred (for put errors)
  *   error/message : human-readable error message
  */
-static const char *const path_node = "/path";
-static const char *const info_node = "/lens/info";
-static const char *const id_node = "/lens/id";
+static const char *const s_path = "path";
+static const char *const s_lens = "lens";
+static const char *const s_info = "info";
+static const char *const s_id   = "id";
 
-static const char *const err_node = "/error";
-/* These are all put underneath "/error" */
-static const char *const err_pos_node = "/pos";
-static const char *const err_path_node = "/path";
-static const char *const err_msg_node = "/message";
+static const char *const s_error = "error";
+/* These are all put underneath "error" */
+static const char *const s_pos     = "pos";
+static const char *const s_message = "message";
+static const char *const s_line    = "line";
+static const char *const s_char    = "char";
 
 /*
  * Filters
@@ -91,7 +95,15 @@ static const char *pathbase(const char *path) {
     return (p == NULL) ? path : p + 1;
 }
 
-static int filter_generate(struct filter *filter, const char *root,
+static bool is_excl(struct tree *f) {
+    return streqv(f->label, "excl") && f->value != NULL;
+}
+
+static bool is_incl(struct tree *f) {
+    return streqv(f->label, "incl") && f->value != NULL;
+}
+
+static int filter_generate(struct tree *xfm, const char *root,
                            int *nmatches, char ***matches) {
     glob_t globbuf;
     int gl_flags = glob_flags;
@@ -101,11 +113,11 @@ static int filter_generate(struct filter *filter, const char *root,
     *nmatches = 0;
     *matches = NULL;
 
-    list_for_each(f, filter) {
+    list_for_each(f, xfm->children) {
         char *globpat = NULL;
-        if (! f->include)
+        if (! is_incl(f))
             continue;
-        pathjoin(&globpat, 2, root, f->glob->str);
+        pathjoin(&globpat, 2, root, f->value);
         r = glob(globpat, gl_flags, NULL, &globbuf);
         free(globpat);
 
@@ -121,14 +133,14 @@ static int filter_generate(struct filter *filter, const char *root,
     globbuf.gl_pathv = NULL;
     globbuf.gl_pathc = 0;
 
-    list_for_each(e, filter) {
-        if (e->include)
+    list_for_each(e, xfm->children) {
+        if (! is_excl(e))
             continue;
         for (int i=0; i < pathc;) {
             const char *path = pathv[i];
-            if (strchr(e->glob->str, SEP) == NULL)
+            if (strchr(e->value, SEP) == NULL)
                 path = pathbase(path);
-            if (fnmatch(e->glob->str, path, fnm_flags) == 0) {
+            if (fnmatch(e->value, path, fnm_flags) == 0) {
                 free(pathv[i]);
                 pathc -= 1;
                 if (i < pathc) {
@@ -151,16 +163,18 @@ static int filter_generate(struct filter *filter, const char *root,
     return ret;
 }
 
-static int filter_matches(struct filter *filter, const char *path) {
+static int filter_matches(struct tree *xfm, const char *path) {
     int found = 0;
-    list_for_each(f, filter) {
-        if (f->include)
-            found |= (fnmatch(f->glob->str, path, fnm_flags) == 0);
+    list_for_each(f, xfm->children) {
+        if (is_incl(f) && fnmatch(f->value, path, fnm_flags) == 0) {
+            found = 1;
+            break;
+        }
     }
     if (! found)
         return 0;
-    list_for_each(f, filter) {
-        if (!f->include && (fnmatch(f->glob->str, path, fnm_flags) == 0))
+    list_for_each(f, xfm->children) {
+        if (is_excl(f) && (fnmatch(f->value, path, fnm_flags) == 0))
             return 0;
     }
     return 1;
@@ -188,10 +202,54 @@ void free_transform(struct transform *xform) {
 
 static char *err_path(const char *filename) {
     char *result = NULL;
-    pathjoin(&result, 3, AUGEAS_META_FILES, filename, err_node);
+    if (filename == NULL)
+        pathjoin(&result, 2, AUGEAS_META_FILES, s_error);
+    else
+        pathjoin(&result, 3, AUGEAS_META_FILES, filename, s_error);
     return result;
 }
 
+ATTRIBUTE_FORMAT(printf, 4, 5)
+static int err_set(struct augeas *aug, char **ep, const char *sub,
+                   const char *format, ...) {
+    int r;
+    va_list ap;
+    size_t ep_len = strlen(*ep);
+    char *value = NULL;
+
+    r = pathjoin(ep, 1, sub);
+    if (r < 0)
+        goto error;
+
+    va_start(ap, format);
+    r = vasprintf(&value, format, ap);
+    va_end(ap);
+    if (r < 0) {
+        value = NULL;
+        goto error;
+    }
+
+    r = aug_set(aug, *ep, value);
+ error:
+    free(value);
+    (*ep)[ep_len] = '\0';
+    return (r < 0) ? -1 : 0;
+}
+
+static void
+calc_line_ofs(const char *text, size_t pos, size_t *line, size_t *ofs)
+{
+    *line = 1;
+    *ofs = 0;
+    for (const char *t = text; t < text + pos; t++) {
+        *ofs += 1;
+        if (*t == '\n') {
+            *ofs = 0;
+            *line += 1;
+        }
+    }
+}
+
 /* Record an error in the tree. The error will show up underneath
  * /augeas/FILENAME/error. PATH is the path to the toplevel node in the
  * tree where the lens application happened. When STATUS is NULL, just
@@ -200,15 +258,13 @@ static char *err_path(const char *filename) {
 static int store_error(struct augeas *aug,
                        const char *filename, const char *path,
                        const char *status, int errnum,
-                       const struct lns_error *err) {
+                       const struct lns_error *err, const char *text) {
     char *ep = err_path(filename);
-    size_t ep_len;
     int r;
     int result = -1;
 
     if (ep == NULL)
         return -1;
-    ep_len = strlen(ep);
 
     aug_rm(aug, ep);
     if (status != NULL) {
@@ -216,49 +272,32 @@ static int store_error(struct augeas *aug,
         if (r < 0)
             goto done;
 
+        /* Errors from err_set are ignored on purpose. We try
+         * to report as much as we can */
         if (err != NULL) {
-            if (err->pos > 0) {
-                char *pos;
-
-                r = pathjoin(&ep, 1, err_pos_node);
-                if (r < 0)
-                    goto done;
-                r = asprintf(&pos, "%d", err->pos);
-                if (r < 0)
-                    goto done;
-                r = aug_set(aug, ep, pos);
-                FREE(pos);
-                if (r < 0)
-                    goto done;
-            } else {
-                char *p;
-
-                r = pathjoin(&ep, 1, err_path_node);
-                if (r < 0)
-                    goto done;
-                r = asprintf(&p, "%s%s", path, err->path);
-                if (r < 0)
-                    goto done;
-                r = aug_set(aug, ep, p);
-                FREE(p);
-                if (r < 0)
-                    goto done;
+            if (err->pos >= 0) {
+                size_t line, ofs;
+                err_set(aug, &ep, s_pos, "%d", err->pos);
+                if (text != NULL) {
+                    calc_line_ofs(text, err->pos, &line, &ofs);
+                    err_set(aug, &ep, s_line, "%zd", line);
+                    err_set(aug, &ep, s_char, "%zd", ofs);
+                }
             }
-            ep[ep_len] = '\0';
-            r = pathjoin(&ep, 1, err_msg_node);
-            if (r < 0)
-                goto done;
-            r = aug_set(aug, ep, err->message);
-            if (r < 0)
-                goto done;
+            if (err->path != NULL) {
+                err_set(aug, &ep, s_path, "%s%s", path, err->path);
+            }
+            if (err->lens != NULL) {
+                char *fi = format_info(err->lens->info);
+                if (fi != NULL) {
+                    err_set(aug, &ep, s_lens, "%s", fi);
+                    free(fi);
+                }
+            }
+            err_set(aug, &ep, s_message, "%s", err->message);
         } else if (errnum != 0) {
             const char *msg = strerror(errnum);
-            r = pathjoin(&ep, 1, err_msg_node);
-            if (r < 0)
-                goto done;
-            r = aug_set(aug, ep, msg);
-            if (r < 0)
-                goto done;
+            err_set(aug, &ep, s_message, "%s", msg);
         }
     }
 
@@ -281,7 +320,7 @@ static int add_load_info(struct augeas *aug, const char *filename,
         goto done;
     end = strlen(p);
 
-    r = pathjoin(&p, 1, path_node);
+    r = pathjoin(&p, 1, s_path);
     if (r < 0)
         goto done;
 
@@ -290,7 +329,7 @@ static int add_load_info(struct augeas *aug, const char *filename,
         goto done;
     p[end] = '\0';
 
-    r = pathjoin(&p, 1, info_node);
+    r = pathjoin(&p, 2, s_lens, s_info);
     if (r < 0)
         goto done;
 
@@ -303,7 +342,7 @@ static int add_load_info(struct augeas *aug, const char *filename,
         goto done;
     p[end] = '\0';
 
-    r = pathjoin(&p, 1, id_node);
+    r = pathjoin(&p, 2, s_lens, s_id);
     if (r < 0)
         goto done;
 
@@ -372,13 +411,13 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
         goto done;
     }
 
-    aug_tree_replace(aug, path, tree);
+    tree_replace(aug->origin, path, tree);
     tree = NULL;
 
     result = 0;
  done:
     store_error(aug, filename + strlen(aug->root) - 1, path, err_status,
-                errno, err);
+                errno, err, text);
     free_lns_error(err);
     free(path);
     free_tree(tree);
@@ -387,28 +426,110 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
     return result;
 }
 
-int transform_load(struct augeas *aug, struct transform *xform) {
+/* The lens for a transform can be referred to in one of two ways:
+ * either by a fully qualified name "Module.lens" or by the special
+ * syntax "@Module"; the latter means we should take the lens from the
+ * autoload transform for Module
+ */
+static struct lens *lens_from_name(struct augeas *aug, const char *name) {
+    if (name[0] == '@') {
+        struct module *modl = NULL;
+        for (modl = aug->modules;
+             modl != NULL && !streqv(modl->name, name + 1);
+             modl = modl->next);
+        if (modl == NULL || modl->autoload == NULL)
+            return NULL;
+        return modl->autoload->lens;
+    } else {
+        return lens_lookup(aug, name);
+    }
+}
+
+static struct lens *xfm_lens(struct augeas *aug, struct tree *xfm) {
+    struct tree *l = NULL;
+
+    for (l = xfm->children;
+         l != NULL && !streqv("lens", l->label);
+         l = l->next);
+
+    if (l == NULL || l->value == NULL)
+        return NULL;
+    return lens_from_name(aug, l->value);
+}
+
+static void xfm_error(struct tree *xfm, const char *msg) {
+    char *v = strdup(msg);
+    char *l = strdup("error");
+    struct tree *e = NULL;
+
+    if (l == NULL || v == NULL)
+        return;
+    e = tree_append(xfm, l, v);
+}
+
+int transform_validate(struct augeas *aug, struct tree *xfm) {
+    struct tree *l = NULL;
+
+    for (struct tree *t = xfm->children; t != NULL; ) {
+        if (streqv(t->label, "lens"))
+            l = t;
+        if (streqv(t->label, "error")) {
+            struct tree *del = t;
+            t = del->next;
+            tree_unlink(del);
+        } else {
+            t = t->next;
+        }
+    }
+
+    if (l == NULL) {
+        xfm_error(xfm, "missing a child with label 'lens'");
+        return -1;
+    }
+    if (l->value == NULL) {
+        xfm_error(xfm, "the 'lens' node does not contain a lens name");
+        return -1;
+    }
+    if (lens_from_name(aug, l->value) == NULL) {
+        char *msg;
+        if (asprintf(&msg, "the lens '%s' does not exist", l->value) < 0) {
+            xfm_error(xfm, "the lens does not exist");
+        } else {
+            xfm_error(xfm, msg);
+            free(msg);
+        }
+        return -1;
+    }
+    return 0;
+}
+
+int transform_load(struct augeas *aug, struct tree *xfm) {
     int nmatches;
     char **matches;
+    struct lens *lens = xfm_lens(aug, xfm);
     int r;
 
-    r = filter_generate(xform->filter, aug->root, &nmatches, &matches);
+    if (lens == NULL) {
+        // FIXME: Record an error and return 0
+        return -1;
+    }
+    r = filter_generate(xfm, aug->root, &nmatches, &matches);
     if (r == -1)
         return -1;
     for (int i=0; i < nmatches; i++) {
-        load_file(aug, xform->lens, matches[i]);
+        load_file(aug, lens, matches[i]);
         free(matches[i]);
     }
-    lens_release(xform->lens);
+    lens_release(lens);
     free(matches);
     return 0;
 }
 
-int transform_applies(struct transform *xform, const char *path) {
+int transform_applies(struct tree *xfm, const char *path) {
     if (STRNEQLEN(path, AUGEAS_FILES_TREE, strlen(AUGEAS_FILES_TREE))
         || path[strlen(AUGEAS_FILES_TREE)] != SEP)
         return 0;
-    return filter_matches(xform->filter, path + strlen(AUGEAS_FILES_TREE));
+    return filter_matches(xfm, path + strlen(AUGEAS_FILES_TREE));
 }
 
 static int transfer_file_attrs(const char *from, const char *to,
@@ -495,11 +616,19 @@ static int clone_file(const char *from, const char *to,
         *err_status = "clone_read";
         goto done;
     }
-
+    if (fflush(to_fp) != 0) {
+        *err_status = "clone_flush";
+        goto done;
+    }
+    if (fsync(fileno(to_fp)) < 0) {
+        *err_status = "clone_sync";
+        goto done;
+    }
     result = 0;
  done:
     fclose(from_fp);
-    fclose(to_fp);
+    if (fclose(to_fp) != 0)
+        result = -1;
     if (result != 0)
         unlink(to);
     if (result == 0)
@@ -519,6 +648,33 @@ static char *strappend(const char *s1, const char *s2) {
     return result;
 }
 
+static int file_saved_event(struct augeas *aug, const char *path) {
+    const char *saved = strrchr(AUGEAS_EVENTS_SAVED, SEP) + 1;
+    struct pathx *px;
+    struct tree *dummy;
+    int r;
+
+    r = pathx_parse(aug->origin, AUGEAS_EVENTS_SAVED "[last()]",
+                    true, NULL, &px);
+    if (r != PATHX_NOERROR)
+        return -1;
+
+    if (pathx_find_one(px, &dummy) == 1) {
+        r = tree_insert(px, saved, 0);
+        if (r < 0)
+            goto error;
+    }
+
+    if (! tree_set(px, path))
+        goto error;
+
+    free_pathx(px);
+    return 0;
+ error:
+    free_pathx(px);
+    return -1;
+}
+
 /*
  * Save TREE->CHILDREN into the file PATH using the lens from XFORM. Errors
  * are noted in the /augeas/files hierarchy in AUG->ORIGIN under
@@ -532,7 +688,7 @@ static char *strappend(const char *s1, const char *s2) {
  *
  * Return 0 on success, -1 on failure.
  */
-int transform_save(struct augeas *aug, struct transform *xform,
+int transform_save(struct augeas *aug, struct tree *xfm,
                    const char *path, struct tree *tree) {
     FILE *fp = NULL;
     char *augnew = NULL, *augorig = NULL, *augsave = NULL;
@@ -544,8 +700,16 @@ int transform_save(struct augeas *aug, struct transform *xform,
     const char *err_status = NULL;
     char *dyn_err_status = NULL;
     struct lns_error *err = NULL;
+    struct lens *lens = xfm_lens(aug, xfm);
     int result = -1, r;
 
+    errno = 0;
+
+    if (lens == NULL) {
+        err_status = "lens_name";
+        goto done;
+    }
+
     copy_if_rename_fails =
         aug_get(aug, AUGEAS_COPY_IF_RENAME_FAILS, NULL) == 1;
 
@@ -591,16 +755,35 @@ int transform_save(struct augeas *aug, struct transform *xform,
         }
     }
     if (augorig_exists) {
-        if (transfer_file_attrs(augorig_canon, augnew, &err_status) != 0)
+        if (transfer_file_attrs(augorig_canon, augnew, &err_status) != 0) {
+            err_status = "xfer_attrs";
             goto done;
+        }
     }
 
     if (tree != NULL)
-        lns_put(fp, xform->lens, tree->children, text, &err);
-    // FIXME: Delete file if tree == NULL
+        lns_put(fp, lens, tree->children, text, &err);
 
-    if (ferror (fp) || fclose(fp) != 0)
+    if (ferror(fp)) {
+        err_status = "error_augnew";
         goto done;
+    }
+
+    if (fflush(fp) != 0) {
+        err_status = "flush_augnew";
+        goto done;
+    }
+
+    if (fsync(fileno(fp)) < 0) {
+        err_status = "sync_augnew";
+        goto done;
+    }
+
+    if (fclose(fp) != 0) {
+        err_status = "close_augnew";
+        goto done;
+    }
+
     fp = NULL;
 
     if (err != NULL) {
@@ -622,6 +805,10 @@ int transform_save(struct augeas *aug, struct transform *xform,
             result = 0;
             unlink(augnew);
             goto done;
+        } else if (aug->flags & AUG_SAVE_NOOP) {
+            result = 1;
+            unlink(augnew);
+            goto done;
         }
     }
 
@@ -649,13 +836,20 @@ int transform_save(struct augeas *aug, struct transform *xform,
     result = 1;
 
  done:
+    if (result > 0) {
+        r = file_saved_event(aug, path);
+        if (r < 0) {
+            err_status = "saved_event";
+            result = -1;
+        }
+    }
     {
         const char *emsg =
             dyn_err_status == NULL ? err_status : dyn_err_status;
-        store_error(aug, filename, path, emsg, errno, err);
+        store_error(aug, filename, path, emsg, errno, err, NULL);
     }
     free(dyn_err_status);
-    lens_release(xform->lens);
+    lens_release(lens);
     free(text);
     free(augnew);
     if (augorig_canon != augorig)
@@ -669,6 +863,87 @@ int transform_save(struct augeas *aug, struct transform *xform,
     return result;
 }
 
+int remove_file(struct augeas *aug, struct tree *tree) {
+    char *path = NULL;
+    const char *filename = NULL;
+    const char *err_status = NULL;
+    char *dyn_err_status = NULL;
+    char *augsave = NULL, *augorig = NULL, *augorig_canon = NULL;
+    int r;
+
+    path = path_of_tree(tree);
+    if (path == NULL) {
+        err_status = "path_of_tree";
+        goto error;
+    }
+    filename = path + strlen(AUGEAS_META_FILES);
+
+    if ((augorig = strappend(aug->root, filename + 1)) == NULL) {
+        err_status = "root_file";
+        goto error;
+    }
+
+    augorig_canon = canonicalize_file_name(augorig);
+    if (augorig_canon == NULL) {
+        if (errno == ENOENT) {
+            goto done;
+        } else {
+            err_status = "canon_augorig";
+            goto error;
+        }
+    }
+
+    r = file_saved_event(aug, path + strlen(AUGEAS_META_TREE));
+    if (r < 0) {
+        err_status = "saved_event";
+        goto error;
+    }
+
+    if (aug->flags & AUG_SAVE_NOOP)
+        goto done;
+
+    if (aug->flags & AUG_SAVE_BACKUP) {
+        /* Move file to one with extension .augsave */
+        r = asprintf(&augsave, "%s" EXT_AUGSAVE, augorig_canon);
+        if (r == -1) {
+            augsave = NULL;
+                goto error;
+        }
+
+        r = clone_file(augorig_canon, augsave, &err_status, 1);
+        if (r != 0) {
+            dyn_err_status = strappend(err_status, "_augsave");
+            goto error;
+        }
+    } else {
+        /* Unlink file */
+        r = unlink(augorig_canon);
+        if (r < 0) {
+            err_status = "unlink_orig";
+            goto error;
+        }
+    }
+    tree_unlink(tree);
+ done:
+    free(path);
+    free(augorig);
+    free(augorig_canon);
+    free(augsave);
+    return 0;
+ error:
+    {
+        const char *emsg =
+            dyn_err_status == NULL ? err_status : dyn_err_status;
+        store_error(aug, filename, path, emsg, errno, NULL, NULL);
+    }
+    free(path);
+    free(augorig);
+    free(augorig_canon);
+    free(augsave);
+    free(dyn_err_status);
+    return -1;
+}
+
 /*
  * Local variables:
  *  indent-tabs-mode: nil
